/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2006 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.01 of the PHP license,      |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_01.txt                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Mikael Johansson <mikael AT synd DOT info>                   |
  +----------------------------------------------------------------------+
*/

/* $Id: ocal.c,v 1.5 2006/11/08 18:12:10 mikl Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "zend_interfaces.h"
#include "zend_exceptions.h"
#include "php_ocal.h"

/* If you declare any globals in php_ocal.h uncomment this:
ZEND_DECLARE_MODULE_GLOBALS(ocal)
*/

/* True global resources - no need for thread safety here */
static int le_ocal_session;

/* {{{ ocal_functions[]
 */
zend_function_entry ocal_functions[] = {
	{NULL, NULL, NULL}
};
/* }}} */

PHPAPI zend_class_entry *ocal_connection_ce;
PHPAPI zend_class_entry *ocal_sysop_connection_ce;

PHPAPI zend_class_entry *ocal_agenda_interface_ce;
PHPAPI zend_class_entry *ocal_agenda_ce;
PHPAPI zend_class_entry *ocal_composite_agenda_ce;

PHPAPI zend_class_entry *ocal_condition_interface_ce;
PHPAPI zend_class_entry *ocal_condition_ce;
PHPAPI zend_class_entry *ocal_query_ce;

PHPAPI zend_class_entry *ocal_exception_ce;
PHPAPI zend_class_entry *ocal_exception_internal_ce;
PHPAPI zend_class_entry *ocal_exception_argument_ce;

static char **ocal_extract_strings(zval *arr TSRMLS_DC);

/* {{{ ocal_status_exception
   Throws a Zend exception given an Oracle Calendar error code. */
static void ocal_status_exception(char *message, CAPIStatus status TSRMLS_DC)
{
	zend_class_entry *ce;
	char *status_string = NULL;
	CSDK_GetStatusString(status, (const char **) &status_string);

	if (status & CAPI_STATMODE_FATAL) {
		ce = ocal_exception_internal_ce;
	}
	else if (status & CAPI_STAT_DATA) {
		ce = ocal_exception_argument_ce;
	}
	else if (status & CAPI_STAT_SERVICE) {
		ce = ocal_exception_internal_ce;
	}
	else if (status & CAPI_STAT_API) {
		if ((status & CAPI_STAT_API_FLAGS) == CAPI_STAT_API_FLAGS) {
			ce = ocal_exception_argument_ce;
		}
		else {
			ce = ocal_exception_internal_ce;
		}
	}
	else if (status & CAPI_STAT_SECUR) {
		if ((status & CAPI_STAT_SECUR_LOGON) == CAPI_STAT_SECUR_LOGON ||
			(status & CAPI_STAT_SECUR_INSUFFICIENTRIGHTS) == CAPI_STAT_SECUR_INSUFFICIENTRIGHTS) {
			ce = ocal_exception_argument_ce;
		}
		else {
			ce = ocal_exception_internal_ce;
		}
	}
	else if (status & CAPI_STAT_LIBRARY) {
		ce = ocal_exception_internal_ce;
	}
	else if (status & CAPI_STAT_CONFIG) {
		ce = ocal_exception_argument_ce;
	}
	else {
		ce = ocal_exception_ce;
	}

	zend_throw_exception_ex(ce, status TSRMLS_CC, message, status_string);
}
/* }}} */

/* {{{ ocal_session_free
 */
static void ocal_session_free(ocal_session_t *session)
{
	if (session->handle != CSDK_SESSION_INITIALIZER) {
		CSDK_Disconnect(session->handle, CSDK_FLAG_NONE);
		CSDK_DestroySession(&session->handle);
	}
	pefree(session, session->is_persistent);
}
/* }}} */

/* {{{ ocal_session_new
   The session contains the connection handle used in API calls. */
static ocal_session_t *ocal_session_new(int persistent, char *hash, int hash_len, const char *auth, const char *comp, const char *encr TSRMLS_DC)
{
	ocal_session_t *session;
	CAPIStatus status;

	session = (ocal_session_t *) pemalloc(sizeof(ocal_session_t), persistent);
	memset(session, 0, sizeof(*session));

	session->handle = CSDK_SESSION_INITIALIZER;
	session->is_persistent = persistent;

	if ((status = CSDK_CreateSession(CSDK_FLAG_NONE, &session->handle)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not initialize session (%s)", status TSRMLS_CC);
		ocal_session_free(session);
		return NULL;
	}

	if (auth || comp || encr) {
		if ((status = CSDK_ConfigureACE(session->handle, CSDK_FLAG_NONE, auth, comp, encr)) != CAPI_STAT_OK) {
			ocal_status_exception("Could not set authentication/compression/encryption parameters (%s)", status TSRMLS_CC);
			ocal_session_free(session);
			return NULL;
		}
	}

	if (persistent && hash && hash_len) {
		zend_rsrc_list_entry new_le;
		Z_TYPE(new_le) = le_ocal_session;
		new_le.ptr = session;

		if (zend_hash_update(&EG(persistent_list), hash, hash_len+1, (void *) &new_le, sizeof(zend_rsrc_list_entry), NULL) == FAILURE) {
			zend_throw_exception(ocal_exception_internal_ce, "Could not register persistent session", 0 TSRMLS_CC);
			ocal_session_free(session);
			return NULL;
		}

		zend_list_insert(session, le_ocal_session);
	}

	return session;
}
/* }}} */

/* {{{ ocal_session_list_dtor
 */
ZEND_RSRC_DTOR_FUNC(ocal_session_list_dtor)
{
	if (rsrc->ptr) {
		ocal_session_free((ocal_session_t *) rsrc->ptr);
		rsrc->ptr = NULL;
	}
}
/* }}} */

/* {{{ ocal_connection_ping
   Fetches the server version in order to test connectivity. */
static int ocal_session_ping(ocal_session_t *session)
{
	const char *result = NULL;
	if (CSDK_GetCapabilities(session->handle, CAPI_CAPAB_SERVER_VERSION, CSDK_FLAG_NONE, &result) == CAPI_STAT_OK) {
		return 1;
	}
	return 0;
}
/* }}} */

/* {{{ ocal_connection_free_storage
 */
static void ocal_connection_free_storage(ocal_connection_t *connection TSRMLS_DC)
{
	if (connection->session) {
		if (!connection->session->is_persistent) {
			ocal_session_free(connection->session);
		}
		else if (connection->session->handle != CSDK_SESSION_INITIALIZER) {
			CSDK_Deauthenticate(connection->session->handle, CSDK_FLAG_NONE);
		}
	}

	zend_object_std_dtor(&connection->zo TSRMLS_CC);
	efree(connection);
}
/* }}} */

/* {{{ ocal_connection_new
 */
zend_object_value ocal_connection_new(zend_class_entry *ce TSRMLS_DC)
{
	ocal_connection_t *connection;
	zend_object_value retval;
	zval *tmp;

	connection = emalloc(sizeof(*connection));
	memset(connection, 0, sizeof(*connection));

	zend_object_std_init(&connection->zo, ce TSRMLS_CC);
	zend_hash_copy(connection->zo.properties, &ce->default_properties, (copy_ctor_func_t) zval_add_ref, (void *) &tmp, sizeof(zval *));

	retval.handle = zend_objects_store_put(connection, (zend_objects_store_dtor_t) zend_objects_destroy_object, (zend_objects_free_object_storage_t) ocal_connection_free_storage, NULL TSRMLS_CC);
	retval.handlers = &std_object_handlers;

	return retval;
}
/* }}} */

/* {{{ proto void OracleCalendar::__construct(string host, string user, string pass [, bool persistent [, string authentication [, string compression [, string encryption ] ] ] ])
   Oracle Calendar connection. */
static PHP_METHOD(OracleCalendar, __construct)
{
	zval *object = getThis();
	ocal_connection_t *connection;
	ocal_session_t *session = NULL;

	CAPIStatus status;
	char *hash = NULL;
	int hash_len = 0;

	char *host, *user, *pass, *auth = NULL, *comp = NULL, *encr = NULL;
	int host_len, user_len, pass_len, auth_len = 0, comp_len = 0, encr_len = 0;
	zend_bool persistent = 0;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sss|bsss",
		&host, &host_len, &user, &user_len, &pass, &pass_len, &persistent,
		&auth, &auth_len, &comp, &comp_len, &encr, &encr_len) == FAILURE) {
		ZVAL_NULL(object);
		return;
	}

	connection = (ocal_connection_t *) zend_object_store_get_object(object TSRMLS_CC);

	if (persistent) {
		zend_rsrc_list_entry *le;
		hash_len = spprintf(&hash, 0, "ocal:%s:%s:%s:%s", host, auth, comp, encr);

		if (zend_hash_find(&EG(persistent_list), hash, hash_len+1, (void *) &le) != FAILURE &&
			Z_TYPE_P(le) == le_ocal_session && le->ptr && ocal_session_ping((ocal_session_t *) le->ptr)) {
			session = (ocal_session_t *) le->ptr;
		}
	}

	if (!session) {
		if ((session = ocal_session_new(persistent, hash, hash_len, auth_len?auth:NULL, comp_len?comp:NULL, encr_len?encr:NULL TSRMLS_CC)) == NULL) {
			if (hash) {efree(hash);}
			ZVAL_NULL(object);
			return;
		}

		if ((status = CSDK_Connect(session->handle, CSDK_FLAG_NONE, host)) != CAPI_STAT_OK) {
			ocal_status_exception("Could not connect to host (%s)", status TSRMLS_CC);
			ocal_session_free(session);
			if (hash) {efree(hash);}
			ZVAL_NULL(object);
			return;
		}
	}

	if (hash) {
		efree(hash);
	}

	connection->session = session;

	if ((status = CSDK_Authenticate(connection->session->handle, CSDK_FLAG_NONE, user, pass)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not authenticate user (%s)", status TSRMLS_CC);
		ZVAL_NULL(object);
		return;
	}
}
/* }}} */

/* {{{ proto static bool OracleCalendar::setConfigFile(string configfile, string logfile)
   CSDK_SetConfigFile, Allows the SDK to read configuration settings that control error logging and the other configuration parameters. */
static PHP_METHOD(OracleCalendar, setConfigFile)
{
	CAPIStatus status;

	char *configfile, *logfile;
	int configfile_len, logfile_len;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss", &configfile, &configfile_len, &logfile, &logfile_len) == FAILURE) {
		return;
	}

	if ((status = CSDK_SetConfigFile(configfile, logfile)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not set configuration (%s)", status TSRMLS_CC);
		RETURN_FALSE;
	}

	RETURN_TRUE;
}
/* }}} */

/* {{{ proto array OracleCalendar::getCapabilities(int capabilityid)
   CSDK_GetCapabilities, Returns information about this SDK release and/or the Oracle Calendar server. */
static PHP_METHOD(OracleCalendar, getCapabilities)
{
	ocal_connection_t *connection = (ocal_connection_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
	CAPIStatus status;
	const char *result = NULL;

	long capabilityid;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &capabilityid) == FAILURE) {
		return;
	}

	if ((status = CSDK_GetCapabilities(connection->session->handle, (CAPICapabilityID) capabilityid, CSDK_FLAG_NONE, &result)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not get capabilities (%s)", status TSRMLS_CC);
		RETURN_FALSE;
	}

	RETURN_STRING((char *)result, 1)
}
/* }}} */

/* {{{ proto array OracleCalendarAgenda::storeEvents(string events [, int flags ])
   CSDK_StoreEvents, This function reads one VCALENDAR object and stores each contained VEVENT on the server. */
static PHP_METHOD(OracleCalendar, storeEvents)
{
	ocal_connection_t *connection = (ocal_connection_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
    CAPIStream stream = CSDK_STREAM_INITIALIZER;
    CSDKRequestResult result = NULL;
	CAPIHandle handle;
	CAPIStatus status, s;
    const char *uid;
    int i;

	char *ical;
	int ical_len;
    long flags = CSDK_FLAG_STORE_IMPORT;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|l", &ical, &ical_len, &flags) == FAILURE) {
		return;
	}

	if ((status = CSDK_CreateMemoryStream(connection->session->handle, &stream, ical, NULL, CSDK_FLAG_NONE)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not create stream (%s)", status TSRMLS_CC);
		RETURN_FALSE;
	}

	if ((status = CSDK_StoreEvents(connection->session->handle, (CAPIFlag) flags, stream, &result)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not store events (%s)", status TSRMLS_CC);
		CSDK_DestroyStream(connection->session->handle, &stream);
		RETURN_FALSE;
	}

	/* fetch new uids */
	array_init(return_value);
	for (i = 0, s = CSDK_GetFirstResult(result, &handle, &uid, &status); s != CAPI_STAT_DATA_RRESULT_EOR; i++, s = CSDK_GetNextResult(result, &handle, &uid, &status)) {
		add_index_string(return_value, i, (char *) uid, 1);
	}

	CSDK_DestroyResult(&result);
	CSDK_DestroyStream(connection->session->handle, &stream);
}
/* }}} */

/* {{{ proto bool OracleCalendarAgenda::deleteEvents(array uids [, string recurrenceid [, int modifier ] ])
   CSDK_DeleteEvents, Deletes specified events; must be acting as the event owner for this to succeed. */
static PHP_METHOD(OracleCalendar, deleteEvents)
{
	ocal_connection_t *connection = (ocal_connection_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
	CAPIStatus status;
	char **uids;

	zval *zuids;
	char *recurrenceid = NULL;
	int recurrenceid_len = 0;
	long modifier = CAPI_THISINSTANCE;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "a|sl", &zuids, &recurrenceid, &recurrenceid_len, &modifier) == FAILURE) {
		return;
	}

	if ((uids = ocal_extract_strings(zuids TSRMLS_CC)) == NULL) {
		RETURN_FALSE;
	}

	if ((status = CSDK_DeleteEvents(connection->session->handle, CSDK_FLAG_NONE, (CAPIUIDSet) uids, recurrenceid, (int) modifier, NULL)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not delete events (%s)", status TSRMLS_CC);
		RETVAL_FALSE;
	}
	else {
		RETVAL_TRUE;
	}

	efree(uids);
}
/* }}} */

/* {{{ proto array OracleCalendarAgenda::storeTasks(string tasks [, int flags ])
   CSDK_StoreTasks, Creates/modifies tasks on the current user's agenda depending on the store flag passed in. */
static PHP_METHOD(OracleCalendar, storeTasks)
{
	ocal_connection_t *connection = (ocal_connection_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
    CAPIStream stream = CSDK_STREAM_INITIALIZER;
    CSDKRequestResult result = NULL;
	CAPIHandle handle;
	CAPIStatus status, s;
    const char *uid;
    int i;

	char *ical;
	int ical_len;
    long flags = CSDK_FLAG_STORE_IMPORT;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|l", &ical, &ical_len, &flags) == FAILURE) {
		return;
	}

	if ((status = CSDK_CreateMemoryStream(connection->session->handle, &stream, ical, NULL, CSDK_FLAG_NONE)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not create stream (%s)", status TSRMLS_CC);
		RETURN_FALSE;
	}

	if ((status = CSDK_StoreTasks(connection->session->handle, (CAPIFlag) flags, stream, &result)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not store tasks (%s)", status TSRMLS_CC);
		CSDK_DestroyStream(connection->session->handle, &stream);
		RETURN_FALSE;
	}

	/* fetch new uids */
	array_init(return_value);
	for (i = 0, s = CSDK_GetFirstResult(result, &handle, &uid, &status); s != CAPI_STAT_DATA_RRESULT_EOR; i++, s = CSDK_GetNextResult(result, &handle, &uid, &status)) {
		add_index_string(return_value, i, (char *) uid, 1);
	}

	CSDK_DestroyResult(&result);
	CSDK_DestroyStream(connection->session->handle, &stream);
}
/* }}} */

/* {{{ proto bool OracleCalendarAgenda::deleteTasks(array uids)
   CSDK_DeleteTasks, Deletes tasks from the current user's agenda. */
static PHP_METHOD(OracleCalendar, deleteTasks)
{
	ocal_connection_t *connection = (ocal_connection_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
	CAPIStatus status;
	char **uids;

	zval *zuids;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "a", &zuids) == FAILURE) {
		return;
	}

	if ((uids = ocal_extract_strings(zuids TSRMLS_CC)) == NULL) {
		RETURN_FALSE;
	}

	if ((status = CSDK_DeleteTasks(connection->session->handle, CSDK_FLAG_NONE, (CAPIUIDSet) uids, NULL)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not delete tasks (%s)", status TSRMLS_CC);
		RETVAL_FALSE;
	}
	else {
		RETVAL_TRUE;
	}

	efree(uids);
}
/* }}} */

/* {{{ proto array OracleCalendarAgenda::storeContacts(string contacts [, int flags ])
   CSDK_StoreContacts, Stores vCards on the server. */
static PHP_METHOD(OracleCalendar, storeContacts)
{
	ocal_connection_t *connection = (ocal_connection_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
    CAPIStream stream = CSDK_STREAM_INITIALIZER;
    CSDKRequestResult result = NULL;
	CAPIHandle handle;
	CAPIStatus status, s;
    const char *uid;
    int i;

	char *ical;
	int ical_len;
    long flags = CSDK_FLAG_STORE_IMPORT;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|l", &ical, &ical_len, &flags) == FAILURE) {
		return;
	}

	if ((status = CSDK_CreateMemoryStream(connection->session->handle, &stream, ical, NULL, CSDK_FLAG_NONE)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not create stream (%s)", status TSRMLS_CC);
		RETURN_FALSE;
	}

	if ((status = CSDK_StoreContacts(connection->session->handle, (CAPIFlag) flags, stream, &result)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not store contacts (%s)", status TSRMLS_CC);
		CSDK_DestroyStream(connection->session->handle, &stream);
		RETURN_FALSE;
	}

	/* fetch new uids */
	array_init(return_value);
	for (i = 0, s = CSDK_GetFirstResult(result, &handle, &uid, &status); s != CAPI_STAT_DATA_RRESULT_EOR; i++, s = CSDK_GetNextResult(result, &handle, &uid, &status)) {
		add_index_string(return_value, i, (char *) uid, 1);
	}

	CSDK_DestroyResult(&result);
	CSDK_DestroyStream(connection->session->handle, &stream);
}
/* }}} */

/* {{{ proto bool OracleCalendarAgenda::deleteContacts(array uids)
   CSDK_DeleteContacts, Deletes vCards specified by a set of UIDs. */
static PHP_METHOD(OracleCalendar, deleteContacts)
{
	ocal_connection_t *connection = (ocal_connection_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
	CAPIStatus status;
	char **uids;

	zval *zuids;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "a", &zuids) == FAILURE) {
		return;
	}

	if ((uids = ocal_extract_strings(zuids TSRMLS_CC)) == NULL) {
		RETURN_FALSE;
	}

	if ((status = CSDK_DeleteContacts(connection->session->handle, CSDK_FLAG_NONE, (CAPIUIDSet) uids, NULL)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not delete contacts (%s)", status TSRMLS_CC);
		RETVAL_FALSE;
	}
	else {
		RETVAL_TRUE;
	}

	efree(uids);
}
/* }}} */

/* {{{ proto string OracleCalendar::getContactsByQuery(OracleCalendarQuery query [, int flags [, array properties ] ])
   CSDK_FetchContactsByQuery, Fetches vCards which satisfy the conditions specified in the query. */
static PHP_METHOD(OracleCalendar, getContactsByQuery)
{
	ocal_connection_t *connection = (ocal_connection_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
    ocal_query_t *query;
    CAPIStream stream = CSDK_STREAM_INITIALIZER;
	const char *result = NULL;
	CAPIStatus status;

	zval *zquery, *zproperties = NULL;
	long flags = CSDK_FLAG_NONE;
	char **properties = NULL;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "O|la!", &zquery, ocal_query_ce, &flags, &zproperties) == FAILURE) {
		return;
	}

	if ((status = CSDK_CreateMemoryStream(connection->session->handle, &stream, NULL, &result, CSDK_FLAG_NONE)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not create stream (%s)", status TSRMLS_CC);
		RETURN_FALSE;
	}

	query = (ocal_query_t *) zend_object_store_get_object(zquery TSRMLS_CC);

	if (zproperties) {
		properties = ocal_extract_strings(zproperties TSRMLS_CC);
	}

	if ((status = CSDK_FetchContactsByQuery(connection->session->handle, (CAPIFlag) flags, query->handle, (const char **) properties, stream, NULL)) != CAPI_STAT_OK) {
		ocal_status_exception("Could fetch contacts (%s)", status TSRMLS_CC);
		RETVAL_FALSE;
	}
	else {
		RETVAL_STRING((char *) result, 1);
	}

	CSDK_DestroyStream(connection->session->handle, &stream);
	if (properties) {efree(properties);}
}
/* }}} */

/* {{{ proto string OracleCalendar::getContactsByUid(array uids [, int flags [, array properties ] ])
   CSDK_FetchContactsByUID, Fetches vCards by UIDs. */
static PHP_METHOD(OracleCalendar, getContactsByUid)
{
	ocal_connection_t *connection = (ocal_connection_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
    CAPIStream stream = CSDK_STREAM_INITIALIZER;
	const char *result = NULL;
	CAPIStatus status;
	char **uids;

	zval *zuids;
	long flags = CSDK_FLAG_NONE;
	char **properties = NULL;
	zval *zproperties = NULL;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "a|la!", &zuids, &flags, &zproperties) == FAILURE) {
		return;
	}

	if ((status = CSDK_CreateMemoryStream(connection->session->handle, &stream, NULL, &result, CSDK_FLAG_NONE)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not create stream (%s)", status TSRMLS_CC);
		RETURN_FALSE;
	}

	if ((uids = ocal_extract_strings(zuids TSRMLS_CC)) == NULL) {
		CSDK_DestroyStream(connection->session->handle, &stream);
		RETURN_FALSE;
	}

	if (zproperties) {
		properties = ocal_extract_strings(zproperties TSRMLS_CC);
	}

	if ((status = CSDK_FetchContactsByUID(connection->session->handle, (CAPIFlag) flags, (CAPIUIDSet) uids, (const char **) properties, stream, NULL)) != CAPI_STAT_OK) {
		if (!(status & CAPI_STAT_DATA_UID_NOTFOUND)) {
			ocal_status_exception("Could fetch contacts (%s)", status TSRMLS_CC);
		}
		RETVAL_FALSE;
	}
	else {
		RETVAL_STRING((char *) result, 1);
	}

	CSDK_DestroyStream(connection->session->handle, &stream);
	if (properties) {efree(properties);}
	efree(uids);
}
/* }}} */

/* {{{ ocal_sysop_connection_free_storage
 */
static void ocal_sysop_connection_free_storage(ocal_connection_t *connection TSRMLS_DC)
{
	if (connection->session && !connection->session->is_persistent) {
		ocal_session_free(connection->session);
	}

	zend_object_std_dtor(&connection->zo TSRMLS_CC);
	efree(connection);
}
/* }}} */

/* {{{ ocal_sysop_connection_new
 */
zend_object_value ocal_sysop_connection_new(zend_class_entry *ce TSRMLS_DC)
{
	ocal_connection_t *connection;
	zend_object_value retval;
	zval *tmp;

	connection = emalloc(sizeof(*connection));
	memset(connection, 0, sizeof(*connection));

	zend_object_std_init(&connection->zo, ce TSRMLS_CC);
	zend_hash_copy(connection->zo.properties, &ce->default_properties, (copy_ctor_func_t) zval_add_ref, (void *) &tmp, sizeof(zval *));

	retval.handle = zend_objects_store_put(connection, (zend_objects_store_dtor_t) zend_objects_destroy_object, (zend_objects_free_object_storage_t) ocal_sysop_connection_free_storage, NULL TSRMLS_CC);
	retval.handlers = &std_object_handlers;

	return retval;
}
/* }}} */

/* {{{ proto void OracleSysopCalendar::__construct(string host, string nodeid, string pass [, bool persistent [, string authentication [, string compression [, string encryption ] ] ] ])
   SYSOP authenticated calendar connection. */
static PHP_METHOD(OracleSysopCalendar, __construct)
{
	zval *object = getThis();
	ocal_connection_t *connection;
	ocal_session_t *session = NULL;

	CAPIStatus status;
	char *hash = NULL;
	int hash_len = 0;

	char *host, *user, *pass, *auth = NULL, *comp = NULL, *encr = NULL;
	int host_len, user_len, pass_len, auth_len = 0, comp_len = 0, encr_len = 0;
	zend_bool persistent = 0;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sss|bsss",
		&host, &host_len, &user, &user_len, &pass, &pass_len, &persistent,
		&auth, &auth_len, &comp, &comp_len, &encr, &encr_len) == FAILURE) {
		ZVAL_NULL(object);
		return;
	}

	connection = (ocal_connection_t *) zend_object_store_get_object(object TSRMLS_CC);

	if (persistent) {
		zend_rsrc_list_entry *le;
		hash_len = spprintf(&hash, 0, "ocal_sysop:%s:%s:%s:%s:%s:%s", host, user, pass, auth, comp, encr);

		if (zend_hash_find(&EG(persistent_list), hash, hash_len+1, (void *) &le) != FAILURE &&
			Z_TYPE_P(le) == le_ocal_session && le->ptr && ocal_session_ping((ocal_session_t *) le->ptr)) {
			session = (ocal_session_t *) le->ptr;
		}
	}

	if (!session) {
		if ((session = ocal_session_new(persistent, hash, hash_len, auth_len?auth:NULL, comp_len?comp:NULL, encr_len?encr:NULL TSRMLS_CC)) == NULL) {
			if (hash) {efree(hash);}
			ZVAL_NULL(object);
			return;
		}

		if ((status = CSDK_ConnectAsSysop(session->handle, CSDK_FLAG_NONE, host, user, pass)) != CAPI_STAT_OK) {
			ocal_status_exception("Could not connect to host (%s)", status TSRMLS_CC);
			ocal_session_free(session);
			if (hash) {efree(hash);}
			ZVAL_NULL(object);
			return;
		}
	}

	if (hash) {
		efree(hash);
	}

	connection->session = session;
}
/* }}} */

/* {{{ proto bool OracleSysopCalendar::setIdentity(string username)
   CSDK_SetIdentity, Allows an authenticated user to work on behalf of another calendar user or resource. */
static PHP_METHOD(OracleSysopCalendar, setIdentity)
{
	ocal_connection_t *connection = (ocal_connection_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
	CAPIStatus status;

	char *user;
	int user_len;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &user, &user_len) == FAILURE) {
		return;
	}

	if ((status = CSDK_SetIdentity(connection->session->handle, user, CSDK_FLAG_NONE)) != CAPI_STAT_OK) {
		RETURN_FALSE;
	}

	RETURN_TRUE;
}
/* }}} */

/* {{{ ocal_agenda_free_storage
 */
static void ocal_agenda_free_storage(ocal_agenda_t *agenda TSRMLS_DC)
{
	if (agenda->connection) {
		if (agenda->handle != CSDK_HANDLE_INITIALIZER) {
			CSDK_DestroyHandle(agenda->connection->session->handle, &agenda->handle);
		}
		zend_objects_store_del_ref(&agenda->zconnection TSRMLS_CC);
	}

	zend_object_std_dtor(&agenda->zo TSRMLS_CC);
	efree(agenda);
}
/* }}} */

/* {{{ ocal_agenda_new
 */
zend_object_value ocal_agenda_new(zend_class_entry *ce TSRMLS_DC)
{
	ocal_agenda_t *agenda;
	zend_object_value retval;
	zval *tmp;

	agenda = emalloc(sizeof(*agenda));
	memset(agenda, 0, sizeof(*agenda));

	zend_object_std_init(&agenda->zo, ce TSRMLS_CC);
	zend_hash_copy(agenda->zo.properties, &ce->default_properties, (copy_ctor_func_t) zval_add_ref, (void *) &tmp, sizeof(zval *));

	retval.handle = zend_objects_store_put(agenda, (zend_objects_store_dtor_t) zend_objects_destroy_object, (zend_objects_free_object_storage_t) ocal_agenda_free_storage, NULL TSRMLS_CC);
	retval.handlers = &std_object_handlers;

	agenda->handle = CSDK_HANDLE_INITIALIZER;

	return retval;
}
/* }}} */

/* {{{ proto void OracleCalendarAgenda::__construct(OracleCalendar connection, string username)
   An agenda represents a particular users calendar store. */
static PHP_METHOD(OracleCalendarAgenda, __construct)
{
	zval *object = getThis();
	ocal_agenda_t *agenda;
	CAPIStatus status;

	zval *zconnection;
	char *user;
	int user_len;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "Os", &zconnection, ocal_connection_ce, &user, &user_len) == FAILURE) {
		ZVAL_NULL(object);
		return;
	}

	agenda = (ocal_agenda_t *) zend_object_store_get_object(object TSRMLS_CC);

	/* maintain a reference to connection */
	agenda->connection = (ocal_connection_t *) zend_object_store_get_object(zconnection TSRMLS_CC);
	agenda->zconnection = *zconnection;
	zend_objects_store_add_ref(zconnection TSRMLS_CC);

	if ((status = CSDK_GetHandle(agenda->connection->session->handle, user, CSDK_FLAG_NONE, &agenda->handle)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not open agenda (%s)", status TSRMLS_CC);
		ZVAL_NULL(object);
		return;
	}
}
/* }}} */

/* {{{ proto string OracleCalendarAgenda::getName()
   CSDK_GetHandleInfo, Returns the name of the agenda owner, in the form of a sequence of field-value pairs, separated by "/".
   This string, when prepended with a '?' is of an appropriate format to be passed to OracleCalendarAgenda:__construct(). */
static PHP_METHOD(OracleCalendarAgenda, getName)
{
	ocal_agenda_t *agenda = (ocal_agenda_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
	const char *result = NULL;
	CAPIStatus status;

	if ((status = CSDK_GetHandleInfo(agenda->connection->session->handle, agenda->handle, CAPI_HANDLE_NAME, &result)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not fetch agenda info (%s)", status TSRMLS_CC);
		RETURN_FALSE;
	}

	RETURN_STRING((char *) result, 1);
}
/* }}} */

/* {{{ proto string OracleCalendarAgenda::getEmail()
   CSDK_GetHandleInfo, Returns the email address of who the agenda belongs to. */
static PHP_METHOD(OracleCalendarAgenda, getEmail)
{
	ocal_agenda_t *agenda = (ocal_agenda_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
	const char *result = NULL;
	CAPIStatus status;

	if ((status = CSDK_GetHandleInfo(agenda->connection->session->handle, agenda->handle, CAPI_HANDLE_MAILTO, &result)) != CAPI_STAT_OK) {
		if (status != CAPI_STAT_DATA_EMAIL_NOTSET) {
			ocal_status_exception("Could not fetch agenda info (%s)", status TSRMLS_CC);
		}
		RETURN_FALSE;
	}

	RETURN_STRING((char *) result, 1);
}
/* }}} */

/* {{{ ocal_fetch_events_by_range
 */
static void ocal_fetch_events_by_range(INTERNAL_FUNCTION_PARAMETERS, CAPISession session, CAPIHandle *agendas)
{
    CAPIStream stream = CSDK_STREAM_INITIALIZER;
	const char *result = NULL;
	CAPIStatus status;

	char *start, *end;
	int start_len, end_len;
	long flags = CSDK_FLAG_FETCH_COMBINED;
	char **properties = NULL;
	zval *zproperties = NULL;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss|la!", &start, &start_len, &end, &end_len, &flags, &zproperties) == FAILURE) {
		return;
	}

	if ((status = CSDK_CreateMemoryStream(session, &stream, NULL, &result, CSDK_FLAG_NONE)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not create stream (%s)", status TSRMLS_CC);
		RETURN_FALSE;
	}

	if (zproperties) {
		properties = ocal_extract_strings(zproperties TSRMLS_CC);
	}

	if ((status = CSDK_FetchEventsByRange(session, (CAPIFlag) flags, agendas, start, end, (const char **) properties, stream, NULL)) != CAPI_STAT_OK) {
		ocal_status_exception("Could fetch events (%s)", status TSRMLS_CC);
		RETVAL_FALSE;
	}
	else {
		RETVAL_STRING((char *) result, 1);
	}

	CSDK_DestroyStream(session, &stream);
	if (properties) {efree(properties);}
}
/* }}} */

/* {{{ ocal_fetch_events_by_alarm_range
 */
static void ocal_fetch_events_by_alarm_range(INTERNAL_FUNCTION_PARAMETERS, CAPISession session, CAPIHandle *agendas)
{
    CAPIStream stream = CSDK_STREAM_INITIALIZER;
	const char *result = NULL;
	CAPIStatus status;

	char *start, *end;
	int start_len, end_len;
	long flags = CSDK_FLAG_FETCH_COMBINED;
	char **properties = NULL;
	zval *zproperties = NULL;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss|la!", &start, &start_len, &end, &end_len, &flags, &zproperties) == FAILURE) {
		return;
	}

	if ((status = CSDK_CreateMemoryStream(session, &stream, NULL, &result, CSDK_FLAG_NONE)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not create stream (%s)", status TSRMLS_CC);
		RETURN_FALSE;
	}

	if (zproperties) {
		properties = ocal_extract_strings(zproperties TSRMLS_CC);
	}

	if ((status = CSDK_FetchEventsByAlarmRange(session, (CAPIFlag) flags, agendas, start, end, (const char **) properties, stream, NULL)) != CAPI_STAT_OK) {
		ocal_status_exception("Could fetch events (%s)", status TSRMLS_CC);
		RETVAL_FALSE;
	}
	else {
		RETVAL_STRING((char *) result, 1);
	}

	CSDK_DestroyStream(session, &stream);
	if (properties) {efree(properties);}
}
/* }}} */

/* {{{ proto string OracleCalendarAgenda::getEventsByRange(string start, string end [, int flags [, array properties ] ])
   CSDK_FetchEventsByRange, Fetches events which occur within the time range specified. */
static PHP_METHOD(OracleCalendarAgenda, getEventsByRange)
{
	ocal_agenda_t *agenda = (ocal_agenda_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
	CAPIHandle handles[2];

	handles[0] = agenda->handle;
	handles[1] = CSDK_HANDLE_INITIALIZER;

	ocal_fetch_events_by_range(INTERNAL_FUNCTION_PARAM_PASSTHRU, agenda->connection->session->handle, handles);
}
/* }}} */

/* {{{ proto string OracleCalendarAgenda::getEventsByAlarmRange(string start, string end [, int flags [, array properties ] ])
   CSDK_FetchEventsByAlarmRange, Fetches events which have alarms (reminders) that will trigger within the time range specified; the end of the time range is exclusive.*/
static PHP_METHOD(OracleCalendarAgenda, getEventsByAlarmRange)
{
	ocal_agenda_t *agenda = (ocal_agenda_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
	CAPIHandle handles[2];

	handles[0] = agenda->handle;
	handles[1] = CSDK_HANDLE_INITIALIZER;

	ocal_fetch_events_by_alarm_range(INTERNAL_FUNCTION_PARAM_PASSTHRU, agenda->connection->session->handle, handles);
}
/* }}} */

/* {{{ proto string OracleCalendarAgenda::getEventsByUid(array uids [, int flags [, array properties [, string recurrenceid [, int modifier ] ] ] ])
   CSDK_FetchEventsByUID, Fetches events by their UIDs. */
static PHP_METHOD(OracleCalendarAgenda, getEventsByUid)
{
	ocal_agenda_t *agenda = (ocal_agenda_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
    CAPIStream stream = CSDK_STREAM_INITIALIZER;
	const char *result = NULL;
	CAPIStatus status;
	char **uids;

	zval *zuids;
	long flags = CSDK_FLAG_NONE, modifier = CAPI_THISINSTANCE;
	char **properties = NULL, *recurrenceid = NULL;
	zval *zproperties = NULL;
	int recurrenceid_len = 0;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "a|la!sl", &zuids, &flags, &zproperties, &recurrenceid, &recurrenceid_len, &modifier) == FAILURE) {
		return;
	}

	if ((status = CSDK_CreateMemoryStream(agenda->connection->session->handle, &stream, NULL, &result, CSDK_FLAG_NONE)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not create stream (%s)", status TSRMLS_CC);
		RETURN_FALSE;
	}

	if ((uids = ocal_extract_strings(zuids TSRMLS_CC)) == NULL) {
		CSDK_DestroyStream(agenda->connection->session->handle, &stream);
		RETURN_FALSE;
	}

	if (zproperties) {
		properties = ocal_extract_strings(zproperties TSRMLS_CC);
	}

	if ((status = CSDK_FetchEventsByUID(agenda->connection->session->handle, (CAPIFlag) flags, agenda->handle, (CAPIUIDSet) uids, recurrenceid, (int) modifier, (const char **) properties, stream, NULL)) != CAPI_STAT_OK) {
		if (!(status & CAPI_STAT_DATA_UID_NOTFOUND)) {
			ocal_status_exception("Could fetch events (%s)", status TSRMLS_CC);
		}
		RETVAL_FALSE;
	}
	else {
		RETVAL_STRING((char *) result, 1);
	}

	CSDK_DestroyStream(agenda->connection->session->handle, &stream);
	if (properties) {efree(properties);}
	efree(uids);
}
/* }}} */

/* {{{ ocal_fetch_tasks_by_range
 */
static void ocal_fetch_tasks_by_range(INTERNAL_FUNCTION_PARAMETERS, CAPISession session, CAPIHandle *agendas)
{
    CAPIStream stream = CSDK_STREAM_INITIALIZER;
	const char *result = NULL;
	CAPIStatus status;

	char *start, *end;
	int start_len, end_len;
	long flags = CSDK_FLAG_NONE;
	char **properties = NULL;
	zval *zproperties = NULL;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss|la!", &start, &start_len, &end, &end_len, &flags, &zproperties) == FAILURE) {
		return;
	}

	if ((status = CSDK_CreateMemoryStream(session, &stream, NULL, &result, CSDK_FLAG_NONE)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not create stream (%s)", status TSRMLS_CC);
		RETURN_FALSE;
	}

	if (zproperties) {
		properties = ocal_extract_strings(zproperties TSRMLS_CC);
	}

	if ((status = CSDK_FetchTasksByRange(session, (CAPIFlag) flags, agendas, start, end, (const char **) properties, stream, NULL)) != CAPI_STAT_OK) {
		ocal_status_exception("Could fetch tasks (%s)", status TSRMLS_CC);
		RETVAL_FALSE;
	}
	else {
		RETVAL_STRING((char *) result, 1);
	}

	CSDK_DestroyStream(session, &stream);
	if (properties) {efree(properties);}
}
/* }}} */

/* {{{ ocal_fetch_tasks_by_alarm_range
 */
static void ocal_fetch_tasks_by_alarm_range(INTERNAL_FUNCTION_PARAMETERS, CAPISession session, CAPIHandle *agendas)
{
    CAPIStream stream = CSDK_STREAM_INITIALIZER;
	const char *result = NULL;
	CAPIStatus status;

	char *start, *end;
	int start_len, end_len;
	long flags = CSDK_FLAG_NONE;
	char **properties = NULL;
	zval *zproperties = NULL;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss|la!", &start, &start_len, &end, &end_len, &flags, &zproperties) == FAILURE) {
		return;
	}

	if ((status = CSDK_CreateMemoryStream(session, &stream, NULL, &result, CSDK_FLAG_NONE)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not create stream (%s)", status TSRMLS_CC);
		RETURN_FALSE;
	}

	if (zproperties) {
		properties = ocal_extract_strings(zproperties TSRMLS_CC);
	}

	if ((status = CSDK_FetchTasksByAlarmRange(session, (CAPIFlag) flags, agendas, start, end, (const char **) properties, stream, NULL)) != CAPI_STAT_OK) {
		ocal_status_exception("Could fetch tasks (%s)", status TSRMLS_CC);
		RETVAL_FALSE;
	}
	else {
		RETVAL_STRING((char *) result, 1);
	}

	CSDK_DestroyStream(session, &stream);
	if (properties) {efree(properties);}
}
/* }}} */

/* {{{ proto string OracleCalendarAgenda::getTasksByRange(string start, string end [, int flags [, array properties ] ])
   CSDK_FetchTasksByRange, Fetches tasks which are active within the time range specified (the end of the time range is exclusive). */
static PHP_METHOD(OracleCalendarAgenda, getTasksByRange)
{
	ocal_agenda_t *agenda = (ocal_agenda_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
	CAPIHandle handles[2];

	handles[0] = agenda->handle;
	handles[1] = CSDK_HANDLE_INITIALIZER;

	ocal_fetch_tasks_by_range(INTERNAL_FUNCTION_PARAM_PASSTHRU, agenda->connection->session->handle, handles);
}
/* }}} */

/* {{{ proto string OracleCalendarAgenda::getTasksByAlarmRange(string start, string end [, int flags [, array properties ] ])
   CSDK_FetchTasksByAlarmRange, Fetches tasks that have alarms (reminders) that will trigger within the time range specified (the end of the time range is exclusive). */
static PHP_METHOD(OracleCalendarAgenda, getTasksByAlarmRange)
{
	ocal_agenda_t *agenda = (ocal_agenda_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
	CAPIHandle handles[2];

	handles[0] = agenda->handle;
	handles[1] = CSDK_HANDLE_INITIALIZER;

	ocal_fetch_tasks_by_alarm_range(INTERNAL_FUNCTION_PARAM_PASSTHRU, agenda->connection->session->handle, handles);
}
/* }}} */

/* {{{ proto string OracleCalendarAgenda::getTasksByUid(array uids [, int flags [, array properties ] ])
   CSDK_FetchTasksByUID, Fetches tasks by their UIDs. */
static PHP_METHOD(OracleCalendarAgenda, getTasksByUid)
{
	ocal_agenda_t *agenda = (ocal_agenda_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
    CAPIStream stream = CSDK_STREAM_INITIALIZER;
	const char *result = NULL;
	CAPIStatus status;
	char **uids;

	zval *zuids;
	long flags = CSDK_FLAG_NONE;
	char **properties = NULL;
	zval *zproperties = NULL;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "a|la!", &zuids, &flags, &zproperties) == FAILURE) {
		return;
	}

	if ((status = CSDK_CreateMemoryStream(agenda->connection->session->handle, &stream, NULL, &result, CSDK_FLAG_NONE)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not create stream (%s)", status TSRMLS_CC);
		RETURN_FALSE;
	}

	if ((uids = ocal_extract_strings(zuids TSRMLS_CC)) == NULL) {
		CSDK_DestroyStream(agenda->connection->session->handle, &stream);
		RETURN_FALSE;
	}

	if (zproperties) {
		properties = ocal_extract_strings(zproperties TSRMLS_CC);
	}

	if ((status = CSDK_FetchTasksByUID(agenda->connection->session->handle, agenda->handle, (CAPIFlag) flags, (CAPIUIDSet) uids, (const char **) properties, stream, NULL)) != CAPI_STAT_OK) {
		if (!(status & CAPI_STAT_DATA_UID_NOTFOUND)) {
			ocal_status_exception("Could fetch tasks (%s)", status TSRMLS_CC);
		}
		RETVAL_FALSE;
	}
	else {
		RETVAL_STRING((char *) result, 1);
	}

	CSDK_DestroyStream(agenda->connection->session->handle, &stream);
	if (properties) {efree(properties);}
	efree(uids);
}
/* }}} */

/* {{{ ocal_composite_agenda_free_storage
 */
static void ocal_composite_agenda_free_storage(ocal_composite_agenda_t *composite TSRMLS_DC)
{
	int i;

	if (composite->handles) {
		for (i = 0; i<composite->num_handles; i++) {
			zend_objects_store_del_ref(&composite->zagendas[i] TSRMLS_CC);
		}

		efree(composite->zagendas);
		efree(composite->handles);
	}

	if (composite->connection) {
		zend_objects_store_del_ref(&composite->zconnection TSRMLS_CC);
	}

	zend_object_std_dtor(&composite->zo TSRMLS_CC);
	efree(composite);
}
/* }}} */

/* {{{ ocal_composite_agenda_new
 */
zend_object_value ocal_composite_agenda_new(zend_class_entry *ce TSRMLS_DC)
{
	ocal_composite_agenda_t *composite;
	zend_object_value retval;
	zval *tmp;

	composite = emalloc(sizeof(*composite));
	memset(composite, 0, sizeof(*composite));

	zend_object_std_init(&composite->zo, ce TSRMLS_CC);
	zend_hash_copy(composite->zo.properties, &ce->default_properties, (copy_ctor_func_t) zval_add_ref, (void *) &tmp, sizeof(zval *));

	retval.handle = zend_objects_store_put(composite, (zend_objects_store_dtor_t) zend_objects_destroy_object, (zend_objects_free_object_storage_t) ocal_composite_agenda_free_storage, NULL TSRMLS_CC);
	retval.handlers = &std_object_handlers;

	return retval;
}
/* }}} */

/* {{{ proto void OracleCalendarCompositeAgenda:::__construct(OracleCalendar calendar)
   The group agenda construct allows multiple operations to be performed using a single API call. */
static PHP_METHOD(OracleCalendarCompositeAgenda, __construct)
{
	zval *object = getThis();
	ocal_composite_agenda_t *composite;

	zval *zconnection;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "O", &zconnection, ocal_connection_ce) == FAILURE) {
		ZVAL_NULL(object);
		return;
	}

	composite = (ocal_composite_agenda_t *) zend_object_store_get_object(object TSRMLS_CC);

	/* maintain a reference to connection */
	composite->connection = (ocal_connection_t *) zend_object_store_get_object(zconnection TSRMLS_CC);
	composite->zconnection = *zconnection;
	zend_objects_store_add_ref(zconnection TSRMLS_CC);
}
/* }}} */

/* {{{ proto void OracleCalendarCompositeAgenda::addAgenda(OracleCalendarAgenda agenda)
   Adds an agenda to this composite. */
static PHP_METHOD(OracleCalendarCompositeAgenda, addAgenda)
{
	ocal_composite_agenda_t *composite = (ocal_composite_agenda_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
	ocal_agenda_t *agenda;

	zval *zagenda;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "O", &zagenda, ocal_agenda_ce) == FAILURE) {
		return;
	}

	agenda = (ocal_agenda_t *) zend_object_store_get_object(zagenda TSRMLS_CC);

	composite->num_handles++;
	composite->handles = erealloc(composite->handles, sizeof(*composite->handles) * (composite->num_handles + 1));
	composite->handles[composite->num_handles - 1] = agenda->handle;
	composite->handles[composite->num_handles] = CSDK_HANDLE_INITIALIZER;

	/* maintain a reference to agenda */
	composite->zagendas = erealloc(composite->zagendas, sizeof(*composite->zagendas) * composite->num_handles);
	composite->zagendas[composite->num_handles - 1] = *zagenda;
	zend_objects_store_add_ref(zagenda TSRMLS_CC);

	RETURN_TRUE;
}
/* }}} */

/* {{{ proto array OracleCalendarCompositeAgenda:::getAgendas()
   Retrieves the agendas contained in this composite. */
static PHP_METHOD(OracleCalendarCompositeAgenda, getAgendas)
{
	ocal_composite_agenda_t *composite = (ocal_composite_agenda_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
	int i;

	array_init(return_value);
	for (i = 0; i<composite->num_handles; i++) {
		ZVAL_ADDREF(&composite->zagendas[i]);
		add_index_zval(return_value, i, &composite->zagendas[i]);
	}
}
/* }}} */

/* {{{ proto string OracleCalendarCompositeAgenda::getEventsByRange(string start, string end [, int flags [, array properties ] ])
   CSDK_FetchEventsByRange, Fetches events which occur within the time range specified. */
static PHP_METHOD(OracleCalendarCompositeAgenda, getEventsByRange)
{
	ocal_composite_agenda_t *composite = (ocal_composite_agenda_t *) zend_object_store_get_object(getThis() TSRMLS_CC);

	if (!composite->handles) {
		RETURN_FALSE;
	}

	ocal_fetch_events_by_range(INTERNAL_FUNCTION_PARAM_PASSTHRU, composite->connection->session->handle, composite->handles);
}
/* }}} */

/* {{{ proto string OracleCalendarCompositeAgenda::getEventsByAlarmRange(string start, string end [, int flags [, array properties ] ])
   CSDK_FetchEventsByAlarmRange, Fetches events which have alarms (reminders) that will trigger within the time range specified; the end of the time range is exclusive.*/
static PHP_METHOD(OracleCalendarCompositeAgenda, getEventsByAlarmRange)
{
	ocal_composite_agenda_t *composite = (ocal_composite_agenda_t *) zend_object_store_get_object(getThis() TSRMLS_CC);

	if (!composite->handles) {
		RETURN_FALSE;
	}

	ocal_fetch_events_by_alarm_range(INTERNAL_FUNCTION_PARAM_PASSTHRU, composite->connection->session->handle, composite->handles);
}
/* }}} */

/* {{{ proto string OracleCalendarCompositeAgenda::getTasksByRange(string start, string end [, int flags [, array properties ] ])
   CSDK_FetchTasksByRange, Fetches tasks which are active within the time range specified (the end of the time range is exclusive). */
static PHP_METHOD(OracleCalendarCompositeAgenda, getTasksByRange)
{
	ocal_composite_agenda_t *composite = (ocal_composite_agenda_t *) zend_object_store_get_object(getThis() TSRMLS_CC);

	if (!composite->handles) {
		RETURN_FALSE;
	}

	ocal_fetch_tasks_by_range(INTERNAL_FUNCTION_PARAM_PASSTHRU, composite->connection->session->handle, composite->handles);
}
/* }}} */

/* {{{ proto string OracleCalendarCompositeAgenda::getTasksByAlarmRange(string start, string end [, int flags [, array properties ] ])
   CSDK_FetchTasksByAlarmRange, Fetches tasks that have alarms (reminders) that will trigger within the time range specified (the end of the time range is exclusive). */
static PHP_METHOD(OracleCalendarCompositeAgenda, getTasksByAlarmRange)
{
	ocal_composite_agenda_t *composite = (ocal_composite_agenda_t *) zend_object_store_get_object(getThis() TSRMLS_CC);

	if (!composite->handles) {
		RETURN_FALSE;
	}

	ocal_fetch_tasks_by_alarm_range(INTERNAL_FUNCTION_PARAM_PASSTHRU, composite->connection->session->handle, composite->handles);
}
/* }}} */

/* {{{ ocal_query_free_storage
 */
static void ocal_query_free_storage(ocal_query_t *query TSRMLS_DC)
{
	if (query->handle != CSDK_QUERY_INITIALIZER) {
		CSDK_DestroyQuery(&query->handle);
	}

	zend_object_std_dtor(&query->zo TSRMLS_CC);
	efree(query);
}
/* }}} */

/* {{{ ocal_query_new
 */
zend_object_value ocal_query_new(zend_class_entry *ce TSRMLS_DC)
{
	ocal_query_t *query;
	zend_object_value retval;
	zval *tmp;

	query = emalloc(sizeof(*query));
	memset(query, 0, sizeof(*query));

	zend_object_std_init(&query->zo, ce TSRMLS_CC);
	zend_hash_copy(query->zo.properties, &ce->default_properties, (copy_ctor_func_t) zval_add_ref, (void *) &tmp, sizeof(zval *));

	retval.handle = zend_objects_store_put(query, (zend_objects_store_dtor_t) zend_objects_destroy_object, (zend_objects_free_object_storage_t) ocal_query_free_storage, NULL TSRMLS_CC);
	retval.handlers = &std_object_handlers;

	query->handle = CSDK_QUERY_INITIALIZER;

	return retval;
}
/* }}} */

/* {{{ proto void OracleCalendarQuery:::__construct(IOracleCalendarCondition condition)
   Queries are used to search for contacts (vCards.) */
static PHP_METHOD(OracleCalendarQuery, __construct)
{
	zval *object = getThis();
	ocal_query_t *query;
	CSDKCondition cond;
	CAPIStatus status;

	zval *zcondition;
	zval *property, *operator, *value;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "O", &zcondition, ocal_condition_interface_ce) == FAILURE) {
		ZVAL_NULL(object);
		return;
	}

	zend_call_method_with_0_params(&zcondition, Z_OBJCE_P(zcondition), NULL, "getproperty", &property);
	zend_call_method_with_0_params(&zcondition, Z_OBJCE_P(zcondition), NULL, "getoperator", &operator);
	zend_call_method_with_0_params(&zcondition, Z_OBJCE_P(zcondition), NULL, "getvalue", &value);

	cond.prop	= Z_STRVAL_P(property);
	cond.op		= Z_LVAL_P(operator);
	cond.value	= Z_STRVAL_P(value);

	query = (ocal_query_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
	query->handle = CSDK_QUERY_INITIALIZER;

	if ((status = CSDK_CreateQuery(&cond, &query->handle)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not create query (%s)", status TSRMLS_CC);
		ZVAL_NULL(object);
	}

	zval_ptr_dtor(&property);
	zval_ptr_dtor(&operator);
	zval_ptr_dtor(&value);
}
/* }}} */

/* {{{ proto void OracleCalendarQuery::addCondition(IOracleCalendarCondition condition [, int operator ])
   CSDK_AddConditionToQuery, Adds a condition to a query. */
static PHP_METHOD(OracleCalendarQuery, addCondition)
{
	ocal_query_t *query = (ocal_query_t *) zend_object_store_get_object(getThis() TSRMLS_CC);
	CSDKCondition cond;
	CAPIStatus status;

	zval *zcondition;
	long joinop = CSDK_LOP_AND;
	zval *property, *operator, *value;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "O|l", &zcondition, ocal_condition_interface_ce, &joinop) == FAILURE) {
		return;
	}

	zend_call_method_with_0_params(&zcondition, Z_OBJCE_P(zcondition), NULL, "getproperty", &property);
	zend_call_method_with_0_params(&zcondition, Z_OBJCE_P(zcondition), NULL, "getoperator", &operator);
	zend_call_method_with_0_params(&zcondition, Z_OBJCE_P(zcondition), NULL, "getvalue", &value);

	cond.prop	= Z_STRVAL_P(property);
	cond.op		= Z_LVAL_P(operator);
	cond.value	= Z_STRVAL_P(value);

	if ((status = CSDK_AddConditionToQuery(query->handle, &cond, (CSDKOperator) joinop)) != CAPI_STAT_OK) {
		ocal_status_exception("Could not add condition to query (%s)", status TSRMLS_CC);
		RETVAL_FALSE;
	}
	else {
		RETVAL_TRUE;
	}

	zval_ptr_dtor(&property);
	zval_ptr_dtor(&operator);
	zval_ptr_dtor(&value);
}
/* }}} */

/* {{{ proto void OracleCalendarCondition:::__construct(string property, int operator, string value)
   Encapsulates a query conditions such as "FN=Mikael*". */
static PHP_METHOD(OracleCalendarCondition, __construct)
{
	zval *object = getThis();
	zval *property, *operator, *value;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "z/z/z/", &property, &operator, &value) == FAILURE) {
		ZVAL_NULL(object);
		return;
	}

	convert_to_string_ex(&property);
	convert_to_long_ex(&operator);
	convert_to_string_ex(&value);

	zend_update_property(ocal_condition_ce, object, "_property", sizeof("_property")-1, property TSRMLS_CC);
	zend_update_property(ocal_condition_ce, object, "_operator", sizeof("_operator")-1, operator TSRMLS_CC);
	zend_update_property(ocal_condition_ce, object, "_value", sizeof("_value")-1, value TSRMLS_CC);
}
/* }}} */

/* {{{ proto void OracleCalendarCondition:::getProperty()
   Returns the vCard property to match on (eg. FN, NICKNAME, ..) */
static PHP_METHOD(OracleCalendarCondition, getProperty)
{
	zval *result = zend_read_property(ocal_condition_ce, getThis(), "_property", sizeof("_property")-1, 0 TSRMLS_CC);
	RETURN_ZVAL(result, 1, 0);
}
/* }}} */

/* {{{ proto void OracleCalendarCondition:::getOperator()
   Returns the operator to use in the match (eg. AND, OR, ..) */
static PHP_METHOD(OracleCalendarCondition, getOperator)
{
	zval *result = zend_read_property(ocal_condition_ce, getThis(), "_operator", sizeof("_operator")-1, 0 TSRMLS_CC);
	RETURN_ZVAL(result, 1, 0);
}
/* }}} */

/* {{{ proto void OracleCalendarCondition:::getValue()
   Returns the value to match for (eg. "Mikael Johansson".) */
static PHP_METHOD(OracleCalendarCondition, getValue)
{
	zval *result = zend_read_property(ocal_condition_ce, getThis(), "_value", sizeof("_value")-1, 0 TSRMLS_CC);
	RETURN_ZVAL(result, 1, 0);
}
/* }}} */

/* {{{ ocal_extract_strings
   Extracts a PHP array into a null terminated C array, the caller is expected to free the return value. */
static char **ocal_extract_strings(zval *arr TSRMLS_DC)
{
	char **result;
	int count, i = 0;

	HashPosition pos;
    zval **element;

	if (arr == NULL) {
		return NULL;
	}

	convert_to_array_ex(&arr);
	if ((count = zend_hash_num_elements(Z_ARRVAL_P(arr))) == 0) {
		return NULL;
	}

	result = (char **) emalloc(sizeof(char *) * (count + 1));

	for (zend_hash_internal_pointer_reset_ex(Z_ARRVAL_P(arr), &pos);
		 zend_hash_get_current_data_ex(Z_ARRVAL_P(arr), (void **) &element, &pos) == SUCCESS;
		 zend_hash_move_forward_ex(Z_ARRVAL_P(arr), &pos), i++) {
		if (Z_TYPE_PP(element) == IS_STRING) {
			result[i] = Z_STRVAL_PP(element);
		}
		else {
			zval elem = **element;
			zval_copy_ctor(&elem);
			INIT_PZVAL(&elem);
			convert_to_string(&elem);
			result[i] = Z_STRVAL(elem);
			zval_dtor(&elem);
		}
	}

	result[i] = NULL;
	return result;
}
/* }}} */

/* {{{ ocal_connection_funcs[]
 */
static
ZEND_BEGIN_ARG_INFO_EX(arginfo_OracleCalendar___construct, 0, 0, 3)
	ZEND_ARG_INFO(0, host)
	ZEND_ARG_INFO(0, username)
	ZEND_ARG_INFO(0, password)
	ZEND_ARG_INFO(0, persistent)
	ZEND_ARG_INFO(0, authentication)
	ZEND_ARG_INFO(0, compression)
	ZEND_ARG_INFO(0, encryption)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO(arginfo_OracleCalendar_setConfigFile, 0)
	ZEND_ARG_INFO(0, configfile)
	ZEND_ARG_INFO(0, logfile)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO(arginfo_OracleCalendar_getCapabilities, 0)
	ZEND_ARG_INFO(0, capabilityid)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_OracleCalendar_store, 0, 0, 1)
	ZEND_ARG_INFO(0, data)
	ZEND_ARG_INFO(0, flags)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO(arginfo_OracleCalendar_uids, 0)
	ZEND_ARG_ARRAY_INFO(0, uids, 0)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_OracleCalendar_getContactsByQuery, 0, 0, 1)
	ZEND_ARG_OBJ_INFO(0, query, OracleCalendarQuery, 0)
	ZEND_ARG_INFO(0, flags)
	ZEND_ARG_ARRAY_INFO(0, properties, 1)
ZEND_END_ARG_INFO()

static zend_function_entry ocal_connection_funcs[] = {
	PHP_ME(OracleCalendar, __construct,			arginfo_OracleCalendar___construct,			ZEND_ACC_PUBLIC | ZEND_ACC_CTOR)
	PHP_ME(OracleCalendar, setConfigFile,		arginfo_OracleCalendar_setConfigFile,		ZEND_ACC_PUBLIC | ZEND_ACC_STATIC)
	PHP_ME(OracleCalendar, getCapabilities,		arginfo_OracleCalendar_getCapabilities,		ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendar, storeEvents,			arginfo_OracleCalendar_store,				ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendar, deleteEvents,		arginfo_OracleCalendar_uids,				ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendar, storeTasks,			arginfo_OracleCalendar_store,				ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendar, deleteTasks,			arginfo_OracleCalendar_uids,				ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendar, storeContacts,		arginfo_OracleCalendar_store,				ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendar, deleteContacts,		arginfo_OracleCalendar_uids,				ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendar, getContactsByQuery,	arginfo_OracleCalendar_getContactsByQuery,	ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendar, getContactsByUid,	arginfo_OracleCalendar_uids,				ZEND_ACC_PUBLIC)
	{NULL, NULL, NULL}
};
/* }}} */

/* {{{ ocal_sysop_connection_funcs[]
 */
static
ZEND_BEGIN_ARG_INFO_EX(arginfo_OracleSysopCalendar___construct, 0, 0, 3)
	ZEND_ARG_INFO(0, host)
	ZEND_ARG_INFO(0, nodeid)
	ZEND_ARG_INFO(0, password)
	ZEND_ARG_INFO(0, persistent)
	ZEND_ARG_INFO(0, authentication)
	ZEND_ARG_INFO(0, compression)
	ZEND_ARG_INFO(0, encryption)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO(arginfo_OracleSysopCalendar_setIdentity, 0)
	ZEND_ARG_INFO(0, username)
ZEND_END_ARG_INFO()

static zend_function_entry ocal_sysop_connection_funcs[] = {
	PHP_ME(OracleSysopCalendar, __construct,	arginfo_OracleSysopCalendar___construct,	ZEND_ACC_PUBLIC | ZEND_ACC_CTOR)
	PHP_ME(OracleSysopCalendar, setIdentity,	arginfo_OracleSysopCalendar_setIdentity,	ZEND_ACC_PUBLIC)
	{NULL, NULL, NULL}
};
/* }}} */

/* {{{ ocal_agenda_interface_funcs[]
 */
static
ZEND_BEGIN_ARG_INFO_EX(arginfo_IOracleCalendarAgenda_range, 0, 0, 2)
	ZEND_ARG_INFO(0, start)
	ZEND_ARG_INFO(0, end)
	ZEND_ARG_INFO(0, flags)
	ZEND_ARG_ARRAY_INFO(0, properties, 1)
ZEND_END_ARG_INFO()

static zend_function_entry ocal_agenda_interface_funcs[] = {
	PHP_ABSTRACT_ME(IOracleCalendarAgenda, getEventsByRange,		arginfo_IOracleCalendarAgenda_range)
	PHP_ABSTRACT_ME(IOracleCalendarAgenda, getEventsByAlarmRange,	arginfo_IOracleCalendarAgenda_range)
	PHP_ABSTRACT_ME(IOracleCalendarAgenda, getTasksByRange,			arginfo_IOracleCalendarAgenda_range)
	PHP_ABSTRACT_ME(IOracleCalendarAgenda, getTasksByAlarmRange,	arginfo_IOracleCalendarAgenda_range)
	{NULL, NULL, NULL}
};
/* }}} */

/* {{{ ocal_agenda_funcs[]
 */
static
ZEND_BEGIN_ARG_INFO(arginfo_OracleCalendarAgenda___construct, 0)
	ZEND_ARG_OBJ_INFO(0, connection, OracleCalendar, 0)
	ZEND_ARG_INFO(0, username)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_OracleCalendarAgenda_getEventsByUid, 0, 0, 1)
	ZEND_ARG_ARRAY_INFO(0, uids, 0)
	ZEND_ARG_INFO(0, flags)
	ZEND_ARG_ARRAY_INFO(0, properties, 1)
	ZEND_ARG_INFO(0, recurrenceid)
	ZEND_ARG_INFO(0, modifier)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_OracleCalendarAgenda_getTasksByUid, 0, 0, 1)
	ZEND_ARG_ARRAY_INFO(0, uids, 0)
	ZEND_ARG_INFO(0, flags)
	ZEND_ARG_ARRAY_INFO(0, properties, 1)
ZEND_END_ARG_INFO()

static zend_function_entry ocal_agenda_funcs[] = {
	PHP_ME(OracleCalendarAgenda, __construct,			arginfo_OracleCalendarAgenda___construct,		ZEND_ACC_PUBLIC | ZEND_ACC_CTOR)
	PHP_ME(OracleCalendarAgenda, getName,				NULL,											ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendarAgenda, getEmail,				NULL,											ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendarAgenda, getEventsByRange,		arginfo_IOracleCalendarAgenda_range,			ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendarAgenda, getEventsByAlarmRange,	arginfo_IOracleCalendarAgenda_range,			ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendarAgenda, getEventsByUid,		arginfo_OracleCalendarAgenda_getEventsByUid,	ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendarAgenda, getTasksByRange,		arginfo_IOracleCalendarAgenda_range,			ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendarAgenda, getTasksByAlarmRange,	arginfo_IOracleCalendarAgenda_range,			ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendarAgenda, getTasksByUid,			arginfo_OracleCalendarAgenda_getTasksByUid,		ZEND_ACC_PUBLIC)
	{NULL, NULL, NULL}
};
/* }}} */

/* {{{ ocal_composite_agenda_funcs[]
 */
static
ZEND_BEGIN_ARG_INFO(arginfo_OracleCalendarCompositeAgenda___construct, 0)
	ZEND_ARG_OBJ_INFO(0, connection, OracleCalendar, 0)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO(arginfo_OracleCalendarCompositeAgenda_addAgenda, 0)
	ZEND_ARG_OBJ_INFO(0, agenda, OracleCalendarAgenda, 0)
ZEND_END_ARG_INFO()

static zend_function_entry ocal_composite_agenda_funcs[] = {
	PHP_ME(OracleCalendarCompositeAgenda, __construct,				arginfo_OracleCalendarCompositeAgenda___construct,	ZEND_ACC_PUBLIC | ZEND_ACC_CTOR)
	PHP_ME(OracleCalendarCompositeAgenda, addAgenda,				arginfo_OracleCalendarCompositeAgenda_addAgenda,	ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendarCompositeAgenda, getAgendas,				NULL,												ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendarCompositeAgenda, getEventsByRange,			arginfo_IOracleCalendarAgenda_range,				ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendarCompositeAgenda, getEventsByAlarmRange,	arginfo_IOracleCalendarAgenda_range,				ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendarCompositeAgenda, getTasksByRange,			arginfo_IOracleCalendarAgenda_range,				ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendarCompositeAgenda, getTasksByAlarmRange,		arginfo_IOracleCalendarAgenda_range,				ZEND_ACC_PUBLIC)
	{NULL, NULL, NULL}
};
/* }}} */

/* {{{ ocal_condition_interface_funcs[]
 */
static zend_function_entry ocal_condition_interface_funcs[] = {
	PHP_ABSTRACT_ME(IOracleCalendarCondition, getProperty,	NULL)
	PHP_ABSTRACT_ME(IOracleCalendarCondition, getOperator,	NULL)
	PHP_ABSTRACT_ME(IOracleCalendarCondition, getValue,		NULL)
	{NULL, NULL, NULL}
};
/* }}} */

/* {{{ ocal_condition_funcs[]
 */
static
ZEND_BEGIN_ARG_INFO(arginfo_OracleCalendarCondition___construct, 0)
	ZEND_ARG_INFO(0, property)
	ZEND_ARG_INFO(0, operator)
	ZEND_ARG_INFO(0, value)
ZEND_END_ARG_INFO()

static zend_function_entry ocal_condition_funcs[] = {
	PHP_ME(OracleCalendarCondition, __construct,	arginfo_OracleCalendarCondition___construct,	ZEND_ACC_PUBLIC | ZEND_ACC_CTOR)
	PHP_ME(OracleCalendarCondition, getProperty,	NULL,											ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendarCondition, getOperator,	NULL,											ZEND_ACC_PUBLIC)
	PHP_ME(OracleCalendarCondition, getValue,		NULL,											ZEND_ACC_PUBLIC)
	{NULL, NULL, NULL}
};
/* }}} */

/* {{{ ocal_query_funcs[]
 */
static
ZEND_BEGIN_ARG_INFO(arginfo_OracleCalendarQuery___construct, 0)
	ZEND_ARG_OBJ_INFO(0, condition, IOracleCalendarCondition, 0)
ZEND_END_ARG_INFO()

static
ZEND_BEGIN_ARG_INFO_EX(arginfo_OracleCalendarQuery_addCondition, 0, 0, 1)
	ZEND_ARG_OBJ_INFO(0, condition, IOracleCalendarCondition, 0)
	ZEND_ARG_INFO(0, operator)
ZEND_END_ARG_INFO()

static zend_function_entry ocal_query_funcs[] = {
	PHP_ME(OracleCalendarQuery, __construct,	arginfo_OracleCalendarQuery___construct,	ZEND_ACC_PUBLIC | ZEND_ACC_CTOR)
	PHP_ME(OracleCalendarQuery, addCondition,	arginfo_OracleCalendarQuery_addCondition,	ZEND_ACC_PUBLIC)
	{NULL, NULL, NULL}
};
/* }}} */

/* {{{ ocal_module_entry
 */
zend_module_entry ocal_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
	STANDARD_MODULE_HEADER,
#endif
	"ocal",
	ocal_functions,
	PHP_MINIT(ocal),
	PHP_MSHUTDOWN(ocal),
	NULL,
	NULL,
	PHP_MINFO(ocal),
#if ZEND_MODULE_API_NO >= 20010901
	NO_VERSION_YET,			/* Replace with version number for your extension */
#endif
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_OCAL
ZEND_GET_MODULE(ocal)
#endif

/* {{{ PHP_INI
 */
/* Remove comments and fill if you need to have entries in php.ini
PHP_INI_BEGIN()
    STD_PHP_INI_ENTRY("ocal.global_value",      "42", PHP_INI_ALL, OnUpdateLong, global_value, zend_ocal_globals, ocal_globals)
    STD_PHP_INI_ENTRY("ocal.global_string", "foobar", PHP_INI_ALL, OnUpdateString, global_string, zend_ocal_globals, ocal_globals)
PHP_INI_END()
*/
/* }}} */

/* {{{ php_ocal_init_globals
 */
/* Uncomment this function if you have INI entries
static void php_ocal_init_globals(zend_ocal_globals *ocal_globals)
{
	ocal_globals->global_value = 0;
	ocal_globals->global_string = NULL;
}
*/
/* }}} */

#define OCAL_DECLARE_PROPERTY_NULL(ce, name, access_type) \
	zend_declare_property_null(ce, (char *) name, sizeof(name)-1, access_type TSRMLS_CC);

#define OCAL_DECLARE_CLASS_CONST_LONG(ce, name, value) \
	zend_declare_class_constant_long(ce, (char *) name, sizeof(name)-1, (long) value TSRMLS_CC);

#define OCAL_DECLARE_CLASS_CONST_STRING(ce, name, value) \
	zend_declare_class_constant_stringl(ce, (char *) name, sizeof(name)-1, (char *) value, sizeof(value)-1 TSRMLS_CC);

PHP_OCAL_API zend_class_entry *php_ocal_get_exception_base(TSRMLS_D)
{
#if (PHP_MAJOR_VERSION == 5) && (PHP_MINOR_VERSION < 2)
	return zend_exception_get_default();
#else
	return zend_exception_get_default(TSRMLS_C);
#endif
}

/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(ocal)
{
	zend_class_entry ce;

	le_ocal_session = zend_register_list_destructors_ex(NULL, ocal_session_list_dtor, "Oracle Calendar persistent connection", module_number);

	INIT_CLASS_ENTRY(ce, "OracleCalendar", ocal_connection_funcs);
	ocal_connection_ce = zend_register_internal_class(&ce TSRMLS_CC);
	ocal_connection_ce->create_object = ocal_connection_new;

	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"CAPAB_VERSION",					CAPI_CAPAB_VERSION);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"CAPAB_CAPI_VERSION",				CAPI_CAPAB_CAPI_VERSION);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"CAPAB_SERVER_VERSION",				CAPI_CAPAB_SERVER_VERSION);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"CAPAB_MAXDATE",					CAPI_CAPAB_MAXDATE);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"CAPAB_UNSUPPORTED_ICAL_COMP",		CAPI_CAPAB_UNSUPPORTED_ICAL_COMP);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"CAPAB_UNSUPPORTED_ICAL_PROP",		CAPI_CAPAB_UNSUPPORTED_ICAL_PROP);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"CAPAB_AUTH",						CAPI_CAPAB_AUTH);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"CAPAB_COMP",						CAPI_CAPAB_COMP);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"CAPAB_ENCR",						CAPI_CAPAB_ENCR);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"CAPAB_ABOUT_BOX",					CAPI_CAPAB_ABOUT_BOX);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"CAPAB_MINREFRESHRATE",				CAPI_CAPAB_MINREFRESHRATE);

	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"NONE",								CSDK_FLAG_NONE);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"CREATE",							CSDK_FLAG_STORE_CREATE);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"REPLACE",							CSDK_FLAG_STORE_REPLACE);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"IMPORT",							CSDK_FLAG_STORE_IMPORT);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"MODIFY",							CSDK_FLAG_STORE_MODIFY);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"REPLY",							CSDK_FLAG_STORE_REPLY);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"INVITE_SELF",						CSDK_FLAG_STORE_INVITE_SELF);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"STREAM_NOT_MIME",					CSDK_FLAG_STREAM_NOT_MIME);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"NOTIFY_EMAIL",						CAPI_NOTIFY_EMAIL);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"NOTIFY_SMS",						CAPI_NOTIFY_SMS);

	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"THISINSTANCE",						CAPI_THISINSTANCE);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"THISANDPRIOR",						CAPI_THISANDPRIOR);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_connection_ce,	"THISANDFUTURE",					CAPI_THISANDFUTURE);

	INIT_CLASS_ENTRY(ce, "OracleSysopCalendar", ocal_sysop_connection_funcs);
	ocal_sysop_connection_ce = zend_register_internal_class_ex(&ce, ocal_connection_ce, NULL TSRMLS_CC);
	ocal_sysop_connection_ce->create_object = ocal_sysop_connection_new;

	INIT_CLASS_ENTRY(ce, "IOracleCalendarAgenda", ocal_agenda_interface_funcs);
	ocal_agenda_interface_ce = zend_register_internal_interface(&ce TSRMLS_CC);

	INIT_CLASS_ENTRY(ce, "OracleCalendarAgenda", ocal_agenda_funcs);
	ocal_agenda_ce = zend_register_internal_class(&ce TSRMLS_CC);
	ocal_agenda_ce->create_object = ocal_agenda_new;
	zend_class_implements(ocal_agenda_ce TSRMLS_CC, 1, ocal_agenda_interface_ce);

	OCAL_DECLARE_CLASS_CONST_LONG(ocal_agenda_ce,		"AGENDA_ATTENDEE_ONLY",	CSDK_FLAG_FETCH_AGENDA_ATTENDEE_ONLY);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_agenda_ce,		"COMBINED",				CSDK_FLAG_FETCH_COMBINED);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_agenda_ce,		"EXCLUDE_HOLIDAYS",		CSDK_FLAG_FETCH_EXCLUDE_HOLIDAYS);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_agenda_ce,		"EXCLUDE_DAILYNOTES",	CSDK_FLAG_FETCH_EXCLUDE_DAILYNOTES);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_agenda_ce,		"EXCLUDE_DAYEVENTS",	CSDK_FLAG_FETCH_EXCLUDE_DAYEVENTS);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_agenda_ce,		"EXCLUDE_APPOINTMENTS",	CSDK_FLAG_FETCH_EXCLUDE_APPOINTMENTS);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_agenda_ce,		"EXCLUDE_ACCEPTED",		CSDK_FLAG_FETCH_EXCLUDE_ACCEPTED);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_agenda_ce,		"EXCLUDE_DECLINED",		CSDK_FLAG_FETCH_EXCLUDE_DECLINED);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_agenda_ce,		"EXCLUDE_UNCONFIRMED",	CSDK_FLAG_FETCH_EXCLUDE_UNCONFIRMED);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_agenda_ce,		"LOCALTIMES",			CSDK_FLAG_FETCH_LOCALTIMES);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_agenda_ce,		"DO_NOT_EXPAND_RRULE",	CSDK_FLAG_FETCH_DO_NOT_EXPAND_RRULE);

	INIT_CLASS_ENTRY(ce, "OracleCalendarCompositeAgenda", ocal_composite_agenda_funcs);
	ocal_composite_agenda_ce = zend_register_internal_class(&ce TSRMLS_CC);
	ocal_composite_agenda_ce->create_object = ocal_composite_agenda_new;
	zend_class_implements(ocal_composite_agenda_ce TSRMLS_CC, 1, ocal_agenda_interface_ce);

	INIT_CLASS_ENTRY(ce, "IOracleCalendarCondition", ocal_condition_interface_funcs);
	ocal_condition_interface_ce = zend_register_internal_interface(&ce TSRMLS_CC);\

	INIT_CLASS_ENTRY(ce, "OracleCalendarCondition", ocal_condition_funcs);
	ocal_condition_ce = zend_register_internal_class(&ce TSRMLS_CC);
	zend_class_implements(ocal_condition_ce TSRMLS_CC, 1, ocal_condition_interface_ce);

	OCAL_DECLARE_PROPERTY_NULL(ocal_condition_ce,		"_property",			ZEND_ACC_PROTECTED);
	OCAL_DECLARE_PROPERTY_NULL(ocal_condition_ce,		"_operator",			ZEND_ACC_PROTECTED);
	OCAL_DECLARE_PROPERTY_NULL(ocal_condition_ce,		"_value",				ZEND_ACC_PROTECTED);

	OCAL_DECLARE_CLASS_CONST_STRING(ocal_condition_ce,	"FORMATTEDNAME",		CSDK_PROPERTY_FORMATTEDNAME);
	OCAL_DECLARE_CLASS_CONST_STRING(ocal_condition_ce,	"BIRTHDATE",			CSDK_PROPERTY_BIRTHDATE);
	OCAL_DECLARE_CLASS_CONST_STRING(ocal_condition_ce,	"TITLE",				CSDK_PROPERTY_TITLE);
	OCAL_DECLARE_CLASS_CONST_STRING(ocal_condition_ce,	"ROLE",					CSDK_PROPERTY_ROLE);
	OCAL_DECLARE_CLASS_CONST_STRING(ocal_condition_ce,	"COMMENT",				CSDK_PROPERTY_COMMENT);
	OCAL_DECLARE_CLASS_CONST_STRING(ocal_condition_ce,	"CATEGORIES",			CSDK_PROPERTY_CATEGORIES);
	OCAL_DECLARE_CLASS_CONST_STRING(ocal_condition_ce,	"NICKNAME",				CSDK_PROPERTY_NICKNAME);
	OCAL_DECLARE_CLASS_CONST_STRING(ocal_condition_ce,	"SPOUSE",				CSDK_PROPERTY_X_SPOUSE);
	OCAL_DECLARE_CLASS_CONST_STRING(ocal_condition_ce,	"ANNIVERSARY",			CSDK_PROPERTY_X_ANNIVERSARY);
	OCAL_DECLARE_CLASS_CONST_STRING(ocal_condition_ce,	"OFFICE",				CSDK_PROPERTY_X_OFFICE);
	OCAL_DECLARE_CLASS_CONST_STRING(ocal_condition_ce,	"ASSISTANTNAME",		CSDK_PROPERTY_X_ASSISTANTNAME);

	OCAL_DECLARE_CLASS_CONST_LONG(ocal_condition_ce,	"OP_NE",				CSDK_OP_NE);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_condition_ce,	"OP_EQ",				CSDK_OP_EQ);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_condition_ce,	"OP_LT",				CSDK_OP_LT);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_condition_ce,	"OP_LE",				CSDK_OP_LE);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_condition_ce,	"OP_GT",				CSDK_OP_GT);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_condition_ce,	"OP_GE",				CSDK_OP_GE);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_condition_ce,	"OP_STARTSWITH",		CSDK_OP_STARTSWITH);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_condition_ce,	"OP_NE_CS",				CSDK_OP_NE_CS);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_condition_ce,	"OP_EQ_CS",				CSDK_OP_EQ_CS);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_condition_ce,	"OP_STARTSWITH_CS",		CSDK_OP_STARTSWITH_CS);

	INIT_CLASS_ENTRY(ce, "OracleCalendarQuery", ocal_query_funcs);
	ocal_query_ce = zend_register_internal_class(&ce TSRMLS_CC);
	ocal_query_ce->create_object = ocal_query_new;

	OCAL_DECLARE_CLASS_CONST_LONG(ocal_query_ce,		"OP_AND",				CSDK_LOP_AND);
	OCAL_DECLARE_CLASS_CONST_LONG(ocal_query_ce,		"OP_OR",				CSDK_LOP_OR);

	INIT_CLASS_ENTRY(ce, "OracleCalendarException", NULL);
	if ((ocal_exception_ce = zend_register_internal_class_ex(&ce, NULL, "runtimeexception" TSRMLS_CC)) == NULL) {
		ocal_exception_ce = zend_register_internal_class_ex(&ce, php_ocal_get_exception_base(TSRMLS_C), NULL TSRMLS_CC);
	}

	INIT_CLASS_ENTRY(ce, "OracleCalendarInternalException", NULL);
	ocal_exception_internal_ce = zend_register_internal_class_ex(&ce, ocal_exception_ce, NULL TSRMLS_CC);

	INIT_CLASS_ENTRY(ce, "OracleCalendarArgumentException", NULL);
	ocal_exception_argument_ce = zend_register_internal_class_ex(&ce, ocal_exception_ce, NULL TSRMLS_CC);

	/* If you have INI entries, uncomment these lines
	ZEND_INIT_MODULE_GLOBALS(ocal, php_ocal_init_globals, NULL);
	REGISTER_INI_ENTRIES();
	*/
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION
 */
PHP_MSHUTDOWN_FUNCTION(ocal)
{
	/* uncomment this line if you have INI entries
	UNREGISTER_INI_ENTRIES();
	*/
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(ocal)
{
	CAPISession session = CSDK_SESSION_INITIALIZER;
	const char *buf;

	php_info_print_table_start();
	php_info_print_table_header(2, "Oracle Calendar support", "enabled");

	CSDK_GetCapabilities(session, CAPI_CAPAB_VERSION, CSDK_FLAG_NONE, &buf);
	php_info_print_table_row(2, "Calendar SDK version", buf);

	CSDK_GetCapabilities(session, CAPI_CAPAB_MAXDATE, CSDK_FLAG_NONE, &buf);
	php_info_print_table_row(2, "Largest supported date", buf);

	CSDK_GetCapabilities(session, CAPI_CAPAB_UNSUPPORTED_ICAL_COMP, CSDK_FLAG_NONE, &buf);
	php_info_print_table_row(2, "Unsupported iCalendar properties", buf);
	php_info_print_table_end();

	/* Remove comments if you have entries in php.ini
	DISPLAY_INI_ENTRIES();
	*/
}
/* }}} */

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
